# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# import sphinx_rtd_theme

# -- Project information -----------------------------------------------------

project = "echelle"
copyright = "2021, Daniel R Hey"
author = "Daniel R Hey"

# The full version, including alpha/beta/rc tags
# release = "1.4"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    # "sphinx_rtd_theme",
    "nbsphinx",
]
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**/.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "press"
html_theme = "sphinx_book_theme"
html_logo = "logo.png"
# html_theme_options = {
#     "logo_only": False,
#     "display_version": True,
# }



from __future__ import division, print_function

from .echelle import echelle
from .utils import smooth_power

import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
import numpy as np

__all__ = ["interact_echelle"]


def interact_echelle(
    freq,
    power,
    dnu_min,
    dnu_max,
    step=None,
    cmap="BuPu",
    ax=None,
    smooth=False,
    smooth_filter_width=50.0,
    scale=None,
    return_coords=False,
    backend="bokeh",
    notebook_url="localhost:8888",
    plot_method='fast',
    sampling=2,
    **kwargs
):
    """Creates an interactive echelle environment with a variable deltanu
    slider. If you're working in a Jupyter notebook/lab environment, you must
    call `%matplotlib notebook` before running this.

    Parameters
    ----------
    freq : np.array
        Array of frequencies in the amplitude or power spectrum
    power : np.array
        Corresponding array of power values
    dnu_min : float
        Minimum deltanu value for the slider
    dnu_max : float
        Maximum deltanu value for the slider
    step : float, optional
        Step size by which to increment or decrement the slider, by default
        the frequency spacing
    cmap : matplotlib.colormap, optional
        Colormap for the echelle diagram, by default 'BuPu'
    ax : matplotlib.axis, optional
        axis object on which to plot. If none is passed, one will be created,
        by default None
    smooth_filter_width : float, optional
        Size of the Box1DKernel which is convolved with the power to smooth the
        spectrum. 1 performs no smoothing, by default 50.
    scale : str, optional
        either 'sqrt' or 'log' or None. Scales the echelle to bring out more
        features, by default 'sqrt'
    return_coords : bool, optional
        If True, this will bind mouseclick events to the interactive plot.
        Clicking on the plot will store the values of the frequencies
        at the click event, and return them in a list object, by default False
    plot_method : str, 'fast' or 'slow'
        Uses either the fast `pcolormesh` function, or the slower `imshow`
        function. pcolormesh is much faster, but does not support
        interpolation.
    sampling: float
    **kwargs : dict
        Dictionary of arguments to be passed to `echelle.echelle`

    Returns
    -------
    list
        A list of clicked frequencies if `return_coords=True`.
    """

    if dnu_max < dnu_min:
        raise ValueError("Maximum range can not be less than minimum")

    if smooth_filter_width < 1:
        raise ValueError("The smooth filter width can not be less than 1!")

    if smooth:
        power = smooth_power(power, smooth_filter_width)

    x, y, z = echelle(freq, power, (dnu_max + dnu_min) / 2.0, sampling=sampling, **kwargs)

    if scale is "sqrt":
        z = np.sqrt(z)
    elif scale is "log":
        z = np.log10(z)

    if step is None:
        step = 5*np.median(np.diff(freq))
    if backend == "matplotlib":
        # Create the matplotlib version of the interactive
        # form. This should only really be done
        # if the user is working from the terminal.
        if ax is None:
            fig, ax = plt.subplots(figsize=[6.4, 9])
        else:
            fig = plt.gcf()
        
        if plot_method is 'fast':
            line = ax.pcolorfast((x.min(), x.max()),
                                (y.min(), y.max()), 
                                z, 
                                cmap=cmap)
        else:
            line = ax.imshow(
                z,
                aspect="auto",
                extent=(x.min(), x.max(), y.min(), y.max()),
                origin="lower",
                cmap=cmap
            )
        
        axfreq = plt.axes([0.1, 0.025, 0.8, 0.02])
        valfmt = "%1." + str(len(str(step).split(".")[-1])) + "f"
        print(valfmt)
        slider = Slider(
            axfreq,
            u"\u0394\u03BD",
            dnu_min,
            dnu_max,
            valinit=(dnu_max + dnu_min) / 2.0,
            valstep=step,
            valfmt=valfmt,
        )

        def update(dnu):
            x, y, z = echelle(freq, power, dnu, sampling=1, **kwargs)
            if scale is not None:
                if scale is "sqrt":
                    z = np.sqrt(z)
                elif scale is "log":
                    z = np.log10(z)
            line.set_array(z)
            line.set_extent((x.min(), x.max(), y.min(), y.max()))
            ax.set_xlim(0, dnu)
            fig.canvas.blit(ax.bbox)

        def on_key_press(event):
            if event.key == "left":
                new_dnu = slider.val - slider.valstep
            elif event.key == "right":
                new_dnu = slider.val + slider.valstep
            else:
                new_dnu = slider.val

            slider.set_val(new_dnu)
            update(new_dnu)

        def on_click(event):
            ix, iy = event.xdata, event.ydata
            coords.append((ix, iy))

        fig.canvas.mpl_connect("key_press_event", on_key_press)
        slider.on_changed(update)

        ax.set_xlabel(u"Frequency mod \u0394\u03BD")
        ax.set_ylabel("Frequency")
        plt.subplots_adjust(
            left=0.1,
            right=0.95,
            bottom=0.1,
            top=0.95,
        )
        plt.show()

        if return_coords:
            coords = []
            fig.canvas.mpl_connect("button_press_event", on_click)
            return coords

    elif backend == "bokeh":
        # Otherwise we use Bokeh.
        try:
            import bokeh
        except:
            raise ImportError("You need to install the Bokeh package.")

        from bokeh.io import show, output_notebook, push_notebook
        from bokeh.plotting import figure, ColumnDataSource
        from bokeh.layouts import column
        from bokeh.models import CustomJS, ColumnDataSource, FreehandDrawTool
        from bokeh.models import Slider as b_Slider

        import warnings
        from bokeh.util.warnings import BokehUserWarning

        # This is a terrible hack and I hate Bokeh
        warnings.simplefilter("ignore", BokehUserWarning)

        from notebook import notebookapp

        # Make sure the user knows to pass in the right notebook_url.
        servers = list(notebookapp.list_running_servers())
        ports = [s["port"] for s in servers]
        if len(np.unique(ports)) > 1:
            warnings.warn(
                "You have multiple Jupyter servers open. \
            You will need to pass the current notebook to `notebook_url`. \
            i.e. interact_echelle(x,y,notebook_url='http://localhost:8888')",
                UserWarning,
            )

        def create_interact_ui(doc):

            source = ColumnDataSource(
                data={
                    "image": [z],
                    "x": x,
                    "y": y,
                    "dw": [x.max() - x.min()],
                    "dh": [y.max() - y.min()],
                }
            )

            plot = figure(
                x_range=(x.min(), x.max()),
                y_range=(y.min(), y.max()),
                plot_width=550,
                plot_height=600
            )

            palette = get_bokeh_palette(cmap)

            full_plot = plot.image(
                image="image",
                x="x",
                y="y",
                dw="dw",
                dh=y.max() - y.min(),
                source=source,
                palette=palette,
            )

            plot.xaxis.axis_label = u"Frequency mod \u0394\u03BD"
            plot.yaxis.axis_label = "Frequency"

            slider = b_Slider(
                start=dnu_min,
                end=dnu_max,
                value=(dnu_min + dnu_max) / 2,
                step=step,
                title=u"\u0394\u03BD",
                format="0.000",
            )

            # Slider callback
            def update_upon_dnu_change(attr, old, new):
                x, y, z = echelle(
                    freq,
                    power,
                    new,
                    sampling=sampling,
                )
                if scale is not None:
                    if scale is "sqrt":
                        z = np.sqrt(z)
                    elif scale is "log":
                        z = np.log10(z)
                full_plot.data_source.data["image"] = [z]
                full_plot.data_source.data["dw"] = [x.max() - x.min()]
                plot.x_range.start = x.min()
                plot.x_range.end = x.max()

            slider.on_change("value", update_upon_dnu_change)

            # Adjust some toolbar options
            r = plot.multi_line(line_width=15, alpha=0.2, color="red")
            plot.add_tools(FreehandDrawTool(renderers=[r]))
            plot.toolbar.logo = None
            plot.toolbar.active_drag = None

            # Layout all of the plots
            widgets_and_figures = column(slider, plot)
            doc.add_root(widgets_and_figures)

        output_notebook(verbose=False, hide_banner=True)
        return show(create_interact_ui, notebook_url=notebook_url)

    else:
        raise ValueError("'backend' must be either 'matplotlib' or 'bokeh")


def get_bokeh_palette(cmap):
    """Creates a color palette compatible with Bokeh
    from a matplotlib cmap name.

    Parameters
    ----------
    cmap : string
        Name of a matplotlib colormap

    Returns
    -------
    list
        A series of hex colour codes generated from
        the matplotlib colormap
    """
    from bokeh.colors import RGB
    from matplotlib import cm

    # Solution adapted from
    # https://stackoverflow.com/questions/31883097/elegant-way-to-match-a-string-to-a-random-color-matplotlib
    m_RGB = (255 * plt.get_cmap(cmap)(range(256))).astype("int")
    return [RGB(*tuple(rgb)).to_hex() for rgb in m_RGB]


from astropy.convolution import convolve, Box1DKernel


def smooth_power(power, smooth_filter_width):
    """Smooths the input power array with a Box1DKernel from astropy

    Parameters
    ----------
    power : array-like
        Array of power values
    smooth_filter_width : float
        filter width

    Returns
    -------
    array-like
        Smoothed power
    """
    return convolve(power, Box1DKernel(smooth_filter_width))


from __future__ import print_function, division

import warnings

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider

from .utils import smooth_power

__all__ = ["echelle", "plot_echelle"]


def echelle(freq, power, dnu, fmin=0.0, fmax=None, offset=0.0, sampling=0.1):
    """Calculates the echelle diagram. Use this function if you want to do
    some more custom plotting.

    Parameters
    ----------
    freq : array-like
        Frequency values
    power : array-like
        Power values for every frequency
    dnu : float
        Value of deltanu
    fmin : float, optional
        Minimum frequency to calculate the echelle at, by default 0.
    fmax : float, optional
        Maximum frequency to calculate the echelle at. If none is supplied,
        will default to the maximum frequency passed in `freq`, by default None
    offset : float, optional
        An offset to apply to the echelle diagram, by default 0.0

    Returns
    -------
    array-like
        The x, y, and z values of the echelle diagram.
    """
    if fmax is None:
        fmax = freq[-1]

    fmin = fmin - offset
    fmax = fmax - offset
    freq = freq - offset

    if fmin <= 0.0:
        fmin = 0.0
    else:
        fmin = fmin - (fmin % dnu)

    # trim data
    index = (freq >= fmin) & (freq <= fmax)
    trimx = freq[index]

    samplinginterval = np.median(trimx[1:-1] - trimx[0:-2]) * sampling
    xp = np.arange(fmin, fmax + dnu, samplinginterval)
    yp = np.interp(xp, freq, power)

    n_stack = int((fmax - fmin) / dnu)
    n_element = int(dnu / samplinginterval)

    morerow = 2
    arr = np.arange(1, n_stack) * dnu
    arr2 = np.array([arr, arr])
    yn = np.reshape(arr2, len(arr) * 2, order="F")
    yn = np.insert(yn, 0, 0.0)
    yn = np.append(yn, n_stack * dnu) + fmin + offset

    xn = np.arange(1, n_element + 1) / n_element * dnu
    z = np.zeros([n_stack * morerow, n_element])
    for i in range(n_stack):
        for j in range(i * morerow, (i + 1) * morerow):
            z[j, :] = yp[n_element * (i) : n_element * (i + 1)]
    return xn, yn, z


def plot_echelle(
    freq,
    power,
    dnu,
    # mirror=False,
    ax=None,
    cmap="Blues",
    scale=None,
    interpolation=None,
    smooth=False,
    smooth_filter_width=50,
    **kwargs
):
    """Plots the echelle diagram.

    Parameters
    ----------
    freq : numpy array
        Frequency values
    power : array-like
        Power values for every frequency
    dnu : float
        Value of deltanu
    ax : matplotlib.axes._subplots.AxesSubplot, optional
        A matplotlib axes to plot into. If no axes is provided, a new one will
        be generated, by default None
    cmap : str, optional
        A matplotlib colormap, by default 'BuPu'
    scale : str, optional
        either 'sqrt' or 'log' or None. Scales the echelle to bring out more
        features, by default 'sqrt'
    interpolation : str, optional
        Type of interpolation to perform on the echelle diagram through
        matplotlib.pyplot.imshow, by default 'none'
    smooth_filter_width : float, optional
        Amount by which to smooth the power values, using a Box1DKernel
    **kwargs : dict
        Dictionary of arguments to be passed to `echelle.echelle`

    Returns
    -------
    matplotlib.axes._subplots.AxesSubplot
        The plotted echelle diagram on the axes
    """
    if smooth:
        power = smooth_power(power, smooth_filter_width)
    echx, echy, echz = echelle(freq, power, dnu, **kwargs)

    if scale is not None:
        if scale is "log":
            echz = np.log10(echz)
        elif scale is "sqrt":
            echz = np.sqrt(echz)

    if ax is None:
        fig, ax = plt.subplots()

    ax.imshow(
        echz,
        aspect="auto",
        extent=(echx.min(), echx.max(), echy.min(), echy.max()),
        origin="lower",
        cmap=cmap,
        interpolation=interpolation,
    )

    # It's much cheaper just to replot the data we already have
    # and mirror it.
    # if mirror:
    #     ax.imshow(
    #         echz,
    #         aspect="auto",
    #         extent=(
    #             (echx.min() + dnu),
    #             (echx.max() + dnu),
    #             (echy.min() - dnu),
    #             (echy.max()) - dnu,
    #         ),
    #         origin="lower",
    #         cmap=cmap,
    #         interpolation=interpolation,
    #     )

    ax.set_xlabel(r"Frequency" + " mod " + str(dnu))
    ax.set_ylabel(r"Frequency")
    return ax


#!/usr/bin/env python
import os
import sys
from setuptools import setup

# Prepare and send a new release to PyPI
if "release" in sys.argv[-1]:
    os.system("python setup.py sdist")
    os.system("twine upload dist/*")
    os.system("rm -rf dist/echelle*")
    sys.exit()

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

# Load the __version__ variable without importing the package already
exec(open("echelle/version.py").read())

setup(
    name="echelle",
    version=__version__,
    author="Daniel Hey",
    url="https://github.com/danhey/echelle",
    packages=["echelle"],
    description="A Python package for plotting and interacting with echelle diagrams.",
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    install_requires=install_requires,
)


# from matplotlib.testing.decorators import image_comparison

# import echelle
# import numpy as np

# def _run_echelle(pandas=False, N=10000, seed=1234, ndim=3, factor=None,
#                 **kwargs):
#     np.random.seed(seed)
#     data1 = np.random.randn(ndim*4*N//5).reshape([4*N//5, ndim])
#     data2 = (5 * np.random.rand(ndim)[None, :] +
#              np.random.randn(ndim*N//5).reshape([N//5, ndim]))
#     data = np.vstack([data1, data2])
#     if factor is not None:
#         data[:, 0] *= factor
#         data[:, 1] /= factor
#     if pandas:
#         data = pd.DataFrame.from_items(zip(map("d{0}".format, range(ndim)),
#                                            data.T))

#     fig = corner.corner(data, **kwargs)
#     return fig


# # @image_comparison(baseline_images=["basic"], extensions=["png"])
# # def test_basic():
# #     _run_echelle()
